/********************************************************************************************
  BurstCalc.c - burst analysis functions

  2-6-09: In function EstimateBackground, local 'index' variable initialized to 0.  Previously
          causing run-time memory access error.

 ******************************************************************************************/
#include <windows.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Burst.h"

/****************************************************************************************
 Analyzes photon stream to find bursts as defined by the users selected parameters
******************************************************************************************/
int AnalyzeData(PHOTON_STREAM *ps, BURST *burst)
{
  DWORD i, NumBins;
  DWORD Trigger = init.TriggerCounts - 1; 
  
  NumBins = ps->nbins;
  i = 0; 
  burst->numbursts = 0;

  while(i < NumBins) {   // loop over data points
	   if (ps->counts[i] > Trigger) {
		   i = LocateBurst(i, ps, burst);  
		   if(i == burst->dwAnalysisPts) return(BURST_ARRAY_INDEX_OVERFLOW);  // overran burst array size (??)
	   }
	   else 
	     i++; 
  }
   
  // calculate the intervals
  burst->Interval[0] = burst->Time[0];
  for(i=1;i<burst->numbursts;i++) {
         burst->Interval[i] = burst->Time[i] - burst->Time[i-1];   // difference between peak times
  }
  return(NO_BURST_ANALYSIS_ERRORS);
}

/*****************************************************************
  Locates a burst as defined by MinCountsPerBurst and MaxWidth.
  i = "trigger" index of lpData that the burst starts at 
  Initially start is the end of the previous burst
 ****************************************************************/
DWORD LocateBurst(DWORD trigger_index, PHOTON_STREAM *ps, BURST *burst)
{
  DWORD j, k, i1, i2, sum, iTime, count;
  double binwidth = burst->BinWidth;
  DWORD maxbins = (DWORD) ((double) burst->MaxBurstWidth/burst->BinWidth);
  DWORD SubMinBins = burst->SubMinBins;
  DWORD MinCountsPerBurst = burst->MinCountsPerBurst;
  DWORD mincounts = burst->BackGrdCounts + 1;   // smallest acceptable bin value that is in a burst 
  DWORD NumPts = ps->nbins;
  DWORD MaxIndex = NumPts - 1;
  static DWORD old_i2 = 0;

  // step backwards and look for bins > background
  i1 = (trigger_index > 0) ? trigger_index : 1;  // index of trigger bin  
  j = i1 - 1;            // index to start analysis on
  
  // SubMinBins is the number of sequential bins allowed below the threshold (mincounts) within the burst
  if(SubMinBins > 0) {
	  while(1) {   
	      if(j <= old_i2 + SubMinBins || j == 0) break; // don't look before previous burst
	      count = 0;
	      for (k=0;k<SubMinBins;k++)  if(ps->counts[j-k] <= mincounts) count++;
	      if(count == SubMinBins) break;    // no bins above mincounts - leave while loop
  		  i1 = j;             // new presumed start of burst
		  j -= SubMinBins;    // loop back around and check again

	  }  // end of while
     
	  // now go backwards from i1 and find the index of the last bin that counts
      if(i1 > MaxIndex) return(trigger_index + 1);
	  
	  while(ps->counts[i1] > mincounts) {
		 if(i1 > 0) 
			 i1--; 
		 else
			 break; 
	 }
     i1++; 
  }   // end of if statement for SubMinBins > 0
  else  {   // SubMinBins == 0, just back up until first occurence of a bin <= mincounts
        while(1) {
	        if(j <= old_i2) break; // don't look before previous burst
	        if(ps->counts[j] <= mincounts) break;
			i1 = j;
			j--;
		}  // end of while
  }
  
  // i1 should is now set to the start of the presumed burst based on the given parameters
  // Look ahead to see if there are more photons within the burst definition parameters
  
  iTime = i1 + maxbins;               // index of max allowed burst bin (default = 0.5 s max width)
  if(iTime > NumPts) iTime = NumPts; // don't go past the end of the array
    
  // first sum points from before the trigger index
  sum = 0;
  for(j=i1; j<trigger_index; j++) sum += ps->counts[j]; 
  
  // now look after the trigger index
  j = trigger_index;
  i2 = trigger_index + 1;
  if(SubMinBins > 0) {
    while(j < iTime) {
	    count = 0;  
	    for (k=0; k<SubMinBins; k++) if(ps->counts[j+k] <= mincounts) count++;
		if(count == SubMinBins) break ;         // nothing there - leave while loop
		i2 = j;
		j += SubMinBins;
   	}    // end of while
    
	// look ahead from i2 and find the index of the last bin that counts
    while(ps->counts[i2] > mincounts) {
		if(i2 < NumPts) 
			i2++; 
		else
			break;
	}
  }
  else  {          // SubMinBins == 0
        while(j < iTime) {
 	        if(ps->counts[j] <= mincounts)  break; 
		    i2 = j;
			j++;
		}   // end of while
		i2++;
  }    // end of if
  
  // sum the rest of the counts 
  for(j=trigger_index; j<=i2; j++) sum += ps->counts[j];   
  
  // Test to see if the total number of photons is > MinCountsPerBurst:
  if(sum >= MinCountsPerBurst) {                           // a burst, by the user's definition
	  // save number of photons
	  if(init.bBackgroundSubtract) {                       // subtract the average value of the background across the burst 
	      burst->Burst[burst->numbursts] = (double) sum - (burst->BkMean * (double) (i2 - i1)); 
		  if (burst->Burst[burst->numbursts] < 0.0) burst->Burst[burst->numbursts] = 0.0;
	  }
	  else
	      burst->Burst[burst->numbursts] = (double) sum;   
	  
	  burst->Width[burst->numbursts] =  binwidth * (double) (i2 - i1);    // burst width in s
      j = FindPeak(ps->counts, i1, i2);                                   // index of the peak value 
	  burst->Peak[burst->numbursts] = ps->counts[j];  	                  // peak value
	  burst->Time[burst->numbursts] = US_TO_SEC * binwidth * j;           // burst time in data stream in seconds
      burst->Start[burst->numbursts] = i1;                                // start index of burst
	  burst->End[burst->numbursts] = i2;                                  // end index of burst
      burst->numbursts++;                                                 // increment burst counter index
      old_i2 = i2;
	  if(burst->numbursts == burst->dwAnalysisPts) return(burst->dwAnalysisPts); 
      return(i2+1);
  }    
  else {

	  return(trigger_index + 1);  // wasn't a burst
  }
}
 

/******************************************************************************************
   Estimates the background and trigger level for a data set based on the PCH 
 *******************************************************************************************/
void EstimateBackground(PHOTON_STREAM *ps)
{
  DWORD i, j, n;
  DWORD index = 0;
  DWORD *x, *y, ymax; 
  double xmax; 
  char szBuf[32]; 

  if(ps->counts == NULL) return; 

  n = ps->max + 1;                  // 1 greater than max since it is the # of points 
  x = (DWORD *) calloc(n, sizeof(DWORD));
  y = (DWORD *) calloc(n, sizeof(DWORD));

  for(i=0; i<n; i++) x[i] = i;
  
  i = 0;
  while(i < ps->nbins) {    // i is the index of ps->counts
  	   j = 0;           // j is the index of pch->x
	   while(j < n) {
			if(x[j] == (DWORD) ps->counts[i]) {
               y[j] += 1;   
 	           break;   
			}
	        else 
               j++;  // look at next x value
	   }
	   i++;
   }  // end of i < end while
  
  // find the max value and the first value after peak
  ymax = y[0];
  for(i=1;i<n;i++) {
	  if(y[i] > ymax) {
		  ymax = y[i];
		  index = i;
	  }
  }
  xmax = (double) x[index]; 
  init.BackGrdCounts = (WORD) (xmax + sqrt(xmax));   // assume 1 sd from mean
  sprintf(szBuf,"%u", init.BackGrdCounts);
  SendDlgItemMessage(hPSControls, IDC_BACKGRDCOUNTS, WM_SETTEXT,sizeof(szBuf),(LPARAM)((LPSTR) szBuf));
  free(x);
  free(y); 
}


/******************************************************************************************
   Calculate the PCH of the photon stream passed to function. 
   Normally start = 0 and end = nbins-1; 
 *******************************************************************************************/
void CalcPhotonCountHistogram(PHOTON_STREAM *ps, PC_HIST *pch, DWORD start, DWORD end)
{
  DWORD i, j;
  double max, min; 

  end = end + 1;                         // add 1 so it represents the # of pts
  if(end > ps->nbins) end = ps->nbins;   // in case a bad value was passed
  pch->xmax = ps->max;
  pch->n = ps->max + 1;                  // 1 greater than max since it is the # of points 


  if(pch->x != NULL) free(pch->x);
  if(pch->y != NULL) free(pch->y);
  pch->x = (double *) calloc(pch->n, sizeof(double));
  pch->y = (double *) calloc(pch->n, sizeof(double));
  
  strcpy(pch->szFilename, ps->szFilename);
  remove_fext(pch->szFilename);
  
  for(i=0; i<pch->n; i++) pch->x[i] = (double) i;
  
  i = start;
  while(i < end) {    // i is the index of ps->counts
  	   j = 0;           // j is the index of pch->x
	   while(j < pch->n) {
			if(pch->x[j] == (double) ps->counts[i]) {
               pch->y[j] += 1.0;   
 	           break;   
			}
	        else 
               j++;  // look at next xval
	   }
	   i++;
   }  // end of i < end while
  
  GetMinMax(pch->y, pch->n, &max, &min); 
  pch->ymax = max;
}


/****************************************************************************************
  Calculates mean, sd, min and max of array 
  end variable must be 1 less then the array size (i.e. n-1)
  ****************************************************************************************/
 void CalcMean(double *lpData, DWORD start, DWORD end, double *dMean, double *dDev, double *dMin, double *dMax)
{
 DWORD i, n;
 double max, min; 

 n = end - start + 1;  
 max = min = 0; 
 *dMean = *dDev = *dMin = *dMax = 0.0;
 
 for(i=start;i<=end;i++) {
	 *dMean += lpData[i];
	 max = (max > lpData[i]) ? max: lpData[i];
	 min = (min < lpData[i]) ? min: lpData[i];
 }
 *dMax = max;
 *dMin = min; 
 *dMean /= (double) n;
 
 for(i=start;i<=end;i++) *dDev += (*dMean - lpData[i]) * (*dMean - lpData[i]);
 *dDev = sqrt(*dDev /(double) (n));
}

/******************************************************************************************
  Returns the maximum value from a WORD array along with the sum and min of the whole array
  *****************************************************************************************/
WORD GetMaxWord(WORD *a, DWORD n, DWORD *NumCounts, WORD *MinValue)
{
  DWORD i;
  DWORD sum;
  WORD Max, Min;
  
  Max = Min = a[0];
  sum = (DWORD) a[0];
  for (i = 1;i<n;i++)  {
	  if(Max < a[i]) Max = a[i]; 
	  if(Min > a[i]) Min = a[i]; 
	  sum += (DWORD) a[i];
	  }
  *NumCounts = sum;
  *MinValue = Min;
  return(Max);
}

/****************************************************************************
  Returns index of max (peak) value of a WORD array
 ***********************************************************************/
DWORD FindPeak(WORD *a, DWORD start, DWORD end)
{
  DWORD i, index;
  WORD Max = a[start];
  
  index = start; 
  for (i=start+1;i<=end;i++) {
	  if(Max < a[i]) {
		  Max = a[i]; 
		  index = i;
	  }
  }
  return(index);
}

/****************************************************************************
 Returns min and max values of a double array 
 ***********************************************************************/
void GetMinMax(double *a, DWORD n, double *dMin, double *dMax)
{
  DWORD i;
  double min, max; 

  if(a == NULL) return; 
  max = min = a[0];
  for (i=1;i<n;i++) {
	  if(max < a[i]) max = a[i]; 
      if(min > a[i]) min = a[i];
  }
   
  *dMin = min;
  *dMax = max;
}


/****************************************************************************
 Returns the number of values in a WORD array greater than dwVal
 ***********************************************************************/
DWORD NumberBinsAbove(DWORD dwVal, LPWORD a, DWORD n)
{
 DWORD i, dwCounter = 0;

 for(i=0; i<n; i++)  if(a[i] > dwVal) dwCounter++;
 return(dwCounter);
}

/****************************************************************************
  Multiply photon stream by a constant 
  ***********************************************************************/
void MultiplyDataSetByScalar(PHOTON_STREAM *ps, double value)
{
 DWORD i, n;

 n = ps->nbins;
 for(i=0;i<n;i++ ) ps->counts[i] = (WORD) (value * (double) ps->counts[i]);
}

/****************************************************************************
  Add photon stream 2 to photon stream 1
  ***********************************************************************/
void AddDataSets(PHOTON_STREAM *ps1, PHOTON_STREAM *ps2)
{
 DWORD i, n;

 n = ps1->nbins;
 for(i=0;i<n;i++ ) ps1->counts[i] += ps2->counts[i];
}

/****************************************************************************
  Subtract photon stream #2 from photon stream #1
  ***********************************************************************/
void SubtractDataSets(PHOTON_STREAM *ps1, PHOTON_STREAM *ps2)
{
 DWORD i, n;
 int iDiff;
 
 n = ps1->nbins; 
 for(i=0;i<n;i++ ) {
	 iDiff = (int) ps1->counts[i] - (int) ps2->counts[i];
	 if(iDiff < 0) iDiff = 0;
	 ps1->counts[i] = (WORD) iDiff;
 }
}

/****************************************************************************
  Correct photon stream #1 for bleed through from photon stream #2
  ***********************************************************************/
void  BleedThruCorrectDataSets(PHOTON_STREAM *ps1, PHOTON_STREAM *ps2, double frac1, double frac2)
{
 DWORD i, n;
 int ia, ib;
 double a, b, ff;

 n = ps1->nbins; 
 ff = 1 - (frac1 + frac2);
 
 for(i=0;i<n;i++ ) {
	 a = (double) ps1->counts[i];
	 b = (double) ps2->counts[i];
	 ia = (int) ((a - frac2 * (a + b)) / ff);
	 if(ia < 0) ia = 0;
	 ps1->counts[i] =  (WORD) ia;
	 ib = (int) ((b - frac2 * (a + b)) / ff);
	 if(ib < 0) ib = 0;
	 ps2->counts[i] = (WORD) ib; 
 }
}

/********************************************************************************
  Subtract a constant from the photon stream data passed to the function
 ******************************************************************************/
void  SubtractConstantFromDataSet(PHOTON_STREAM *ps, double value)
{
 DWORD i, n;
 double testval;

 n = ps->nbins; 
 for(i=0;i<n;i++ ) {
	 testval = (double) ps->counts[i] - value;
	 if(testval < 0.0) testval = 0.0;
	 ps->counts[i] =  (WORD) testval;
	}
}

/********************************************************************************
  Add a constant to the photon stream data passed to the function
 ******************************************************************************/
void  AddConstantToDataSet(PHOTON_STREAM *ps, double value)
{
 DWORD i, n;

 n = ps->nbins; 
 for(i=0;i<n;i++) ps->counts[i] +=  (WORD) value;
}

/********************************************************************************
  Adds a Poission deviate to the count value - used with synthetic testing data
  ******************************************************************************/
void AddPoisNoiseToDataSet(PHOTON_STREAM *ps)
{
 DWORD i;

 for(i=0;i<ps->nbins;i++ ) {
	 if(ps->counts[i] == 0) {
        ps->counts[i] =+ (WORD) PoisDev(1.0);	 
	 }
	 else {
	   ps->counts[i] += (WORD) PoisDev((double) ps->counts[i]);
	 }
 }
}


/******************************************************************************************************************************
  Rebin data to new binwidth
 ***************************************************************************************************************************/
void ReBinData(PHOTON_STREAM *ps, double NewBinWidth)
{
  
}



