/**********************************************************************************************
  BurstCorCom.c   - I/O routines for Corelator.Com data

  3-17-09: In LoadCorrComFile() func, the incoming variable 'szFilename', 
           which is now the full-path filename, is no longer copied to the 
           PHOTON_STREAM struct's current base filename. 
 *********************************************************************************************/

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "resource.h"
#include "burst.h"

/*
Flex photon trace files:
  
The raw data files record the difference in system clock ticks between photon event.
BYTE 1 =  8 if 8 bit (BYTE);  16 if 16 bit (WORD)
BYTE 2 indentifies the system clock: 40 for 40 MHz or 60 for 60MHz.  
The min time bin is therefore 1/system clock 

The data follows, so raw data size is FileSize - 2 bytes. 

If WORD format, each WORD represents a photon event and time elasped between 
incoming photons= WORD/system clock. If the value is 0xFFFF the next four 
bytes represent a photon event.   

If BYTE format, each BYTE represents a photon event unless the value is 0xFF, in which case, 
the BYTE means 255 clock ticks passed without a photon event. 
For example 0A 0B FF 08 means there are three photon events.  
The time series are 0x0A+1, 0x0B+1, 0xFF+8+1.  
*/

/********************************************************************************
 Loads correlator dot com card's photon history file into a PHOTON struct and bins the data
  **********************************************************************************/ 
int LoadCorrComFile(char *szFilename, PHOTON_STREAM *ps, double binwidth) 
{
  DWORD   i, j, num, n, NumCounts;
  DWORD   dwFileSizeHi, dwNumBytesRead, dwFileSize;
  DWORD   dwSum, m, numbins;
  WORD    Min; 
  LPVOID  lpBuffer = NULL;
  LPBYTE  lpB = NULL;
  LPDWORD lpDW = NULL; 
  HANDLE  hFile;
  BYTE    bClock, bLength;

  // open file for data read
  hFile = CreateFile(szFilename, GENERIC_READ, 0,NULL,OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile == INVALID_HANDLE_VALUE) return(FILE_OPEN_ERROR);
  
  // get virtual mem
  dwFileSize = GetFileSize(hFile, &dwFileSizeHi);
  lpB = (BYTE *) GetVirtualMem(dwFileSize); 
  if(lpB == NULL) return(MEM_ERROR);
  
  // read data
  ReadFile(hFile, lpB, dwFileSize, &dwNumBytesRead, NULL);
  if(dwNumBytesRead == 0) return(ZERO_BYTES_READ); 
  CloseHandle(hFile);

  // read header
  bLength = lpB[0];
  bClock =  lpB[1];
  ps->BaseBinWidth = 1.0/(double) bClock;           // minimum "bin" time in s
  ps->BinWidth = binwidth; 
  m = (DWORD) ((double) bClock * binwidth);    // number of clock ticks per bin 

  if(bLength == 8) { // byte data 
  	  num = dwFileSize - 2;       // max possible number of photons in file
	  
	  // get memory for index array
	  lpDW = (DWORD *) GetVirtualMem(num * sizeof(DWORD));
      if(lpDW == NULL) return(MEM_ERROR);
	
	  // load the index array find the number of fundamental bins
	  // this creates and array (lpDW) whose elements contain the number of ticks between each photon
	  dwSum = 0;      // summing variable for number of fundamental time bins
	  j = 2;          // 2 to offset for 2 byte header
	  i =  n = 0;
	  while(j < dwFileSize) {
		  while(lpB[j] == 0xFF)  {   
			  dwSum += lpB[j];      // 255 clk ticks with no photons
  		      j++;
		  }
	      dwSum += lpB[j];          // add # of ticks to next photons
		  j++;
		  lpDW[i] = dwSum;          // dwSum  = running sum of number of clock ticks
		  n++;    // increment photon counter
		  i++;    // increment lpDW index
	  }
      ps->NumPhotons = n;              // total number of photons found
	  numbins = 1 + dwSum / m;             // total number of clock ticks / number of ticks per bin = # bins
	  ps->nbins = numbins;             // save numbin value           
	  ps->TraceLength = ps->BaseBinWidth * dwSum * 1.0E-6;   // total tracelength in seconds
	  ps->CPS =  (double) (ps->NumPhotons / ps->TraceLength) / 1000.0 ;    // CPS in KHz

      // debug code
	  /*
	  {
		  FILE *fp = fopen("lpdw_dump.txt", "wt");
		  for(i=0;i<n;i++) {
			  fprintf(fp, "%ld\t%ld\t%f\n", lpDW[i], lpDW[i]/m, (double) lpDW[i] / (double) m);
		  }
		  fclose(fp);
	  }
      */

	  // get memory for rebinning array
	  ps->counts = (WORD *) GetVirtualMem(numbins * sizeof(WORD));
      if(ps->counts == NULL) {
		 VirtualFree(lpB, 0, MEM_RELEASE);
         VirtualFree(lpDW, 0, MEM_RELEASE);
		 return(MEM_ERROR);
	  }
	  
      // loop through index and fill count array
	  j = i = 0;
	  while(i<n) {
		  j = lpDW[i] / m;
		  ps->counts[j]++;
		  i++;
	  }
  }  // end of byte data loop
  else  {  // WORD file -- fix later
	  MessageBox(NULL, "Cor.Com file was WORD data -- save as BYTE.", "Load Error", MB_OK); 
	  return(USER_CANCEL);    
  }
  ps->max =  GetMaxWord(ps->counts, numbins, &NumCounts, &Min);
  ps->min = Min;
  ps->SGFilterPasses = 0;
  //strcpy(ps->szFilename, szFilename);    // 3-17-09

  // free mem
  VirtualFree(lpB, 0, MEM_RELEASE);
  VirtualFree(lpDW, 0, MEM_RELEASE);
  return(FALSE);
}


