/*****************************************************************************
   BURSTFCS.C - FCS & PCH functions

   3-17-09: In AutocorrelateSegments() func, the PHOTON_STREAM struct's current
            full-path filename is now copied to local variable 'szBuf'.

            In AutoCorrelate() func, the PHOTON_STREAM struct's current
            full-path filename is now copied to the AUTOCORRELATION struct's
            'szFilename' variable.

            In AutoCorrelate() func, the PHOTON_STREAM struct's current full-path
            filename is now copied to the local variable 'szBuf'.

            In CrossCorrelate() func, the PHOTON_STREAM struct's current
            full-path filename is now copied to the AUTOCORRELATION struct's
            'szFilename' variable.

            In CrossCorrelate() func, the PHOTON_STREAM struct's current
            full-path filename is now copied to the local variable 'szBuf'.
 ***************************************************************************/
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "resource.h"
#include "burst.h"

/***************************************************************************** 
  Multi-Tau scheme:
  M =  number points in trace 
  T = fundamental binsize
  MT = trace length
  s = init.SumSteps
  f = fraction of MT to autocorrelate to
  
	binsize:                    lags:
__________________________________________________
      
	 T                 T       to      2s * T
	2 * T       (s+1) * 2 * T  to  (2s+1) * 2 * T
    4 * T       (s+1) * 4 * T  to  (2s+1) * 4 * T
      	                         
	                         
   2^k * T      (s+1) * s^k * T to (2s+1) * 2^k * T
___________________________________________________
 
   k = floor(ln(M * f/2s)/ln2) for autocorrelation up to f * MT
   default f = 0.5 so k = floor(ln(M/4s)/ln2)
   
   the number of lag values is k(s+1) + 2s

*************************************************************************************/

/******************************************************************************************
  Called by WinProc - sets up for autocorrelation and plots results
 *******************************************************************************************/
void AutoCorrelate(HWND hwnd, int set) 
{
 double CalcTime, StartTime, EndTime; 
 char   szBuf[256];
 int    result, start, end; 
 PHOTON_STREAM ps, *ps_current; 
 AUTOCORRELATION *fcs = GETFCS(set);
 
 
 if(!flags.bSet[set]) {
     MessageBox(hwnd,"No data loaded", "Analyze Error",MB_OK);
	 return;
 }

 if(init.bShowFCSDialog)  result = DialogBox(hInst,"FCSSETUP", hwnd, (DLGPROC)FCSSetupProc); 
 if(result) return;         // user clicked cancel

 HOURGLASS_CURSOR;
 fcs->DataType = cd.DataType = FCS_DATA;
 fcs->nave = 1; 
 TurnOffBackGroundSelect(hwnd);
 if(init.BinWidth == init.FCSBinWidth) {    // use existing binwidth
	ps_current  = GETPS(set); 
    fcs->burst_binwidth = ps_current->BinWidth;
	fcs->binwidth = init.FCSBinWidth; 
	fcs->TraceLength = cd.TraceLength; 
    strcpy(fcs->szFilename, ps_current->szFilePath);   // 3-17-09
	start = ps_current->ACindex1;
	end = ps_current->ACindex2; 
	GetCorrelationParameters(fcs, start, end, init.SumStep);
	GetCorrelationMemory(fcs);
    CalcTime = AutoCorrelateMultiTau(ps_current, fcs, start, end); 
 }
 else {                                    // reload data and autocorrelate at different binwidth
    ps_current = GETPS(set);     // get the filename of the open PS
    strcpy(szBuf, ps_current->szFilePath);    // 3-17-09
    fcs->burst_binwidth = ps_current->BinWidth;
	strcpy(fcs->szFilename, ps_current->szFilePath);   // 3-17-09 
	fcs->binwidth = init.FCSBinWidth; 
	fcs->TraceLength = cd.TraceLength; 
	result = LoadPSWithNewBinWidth(szBuf, &ps, init.FCSBinWidth) ;
	if(result) return;   // file error
	start = GetPSIndex(ps_current->ACindex1, ps_current, &ps); 
	end = GetPSIndex(ps_current->ACindex2, ps_current, &ps); 
	GetCorrelationParameters(fcs, start, end, init.SumStep);
	GetCorrelationMemory(fcs);
    CalcTime = AutoCorrelateMultiTau(&ps, fcs, start, end); 
    FreePhotonStreamMem(&ps); 
 }

 StartTime = (double) start * init.FCSBinWidth * US_TO_SEC;
 EndTime = (double) end * init.FCSBinWidth * US_TO_SEC;
 sprintf(szBuf, "AC from %g to %g s,  MT step = %d (calc:%3.2f sec.)", StartTime, EndTime, init.SumStep, CalcTime); 
 PlotFCSCurve(fcs, szBuf);
 if(!bShowResultsWin) ToggleResultsWindow(hwnd)	 ; 
 ARROW_CURSOR;					  
 }



/******************************************************************************************
  Called by WinProc - sets up for cross-correlation and plots results
 *******************************************************************************************/
void CrossCorrelate(HWND hwnd, int set)
{
 double CalcTime, StartTime, EndTime;
 char   szBuf[256];
 char szFilename[256];
 int    result, start, end;  
 PHOTON_STREAM psA, psB, *ps1_current, *ps2_current;
 AUTOCORRELATION *fcs = GETFCS(set);
 
 if(!flags.bSet[0] || !flags.bSet[1]) {
     MessageBox(hwnd,"Both PS1 and PS2 needed.", "Analyze Error",MB_OK);
	 return;
 }

 if(init.bShowFCSDialog)  result = DialogBox(hInst,"FCSSETUP", hwnd, (DLGPROC)FCSSetupProc); 
 if(result) return;   // user clicked cancel

 HOURGLASS_CURSOR;
 fcs->DataType = cd.DataType = FCCS_DATA;
 fcs->nave = 1;

 TurnOffBackGroundSelect(hwnd);
 if(init.BinWidth == init.FCSBinWidth) {    // use existing binwidth
	ps1_current = GETPS(set); 
	ps2_current = GETPS(GETOTHER(set));
	start = ps1_current->ACindex1;
	end = ps1_current->ACindex2;
    strcpy(fcs->szFilename, ps1_current->szFilePath);    // 3-17-09
	fcs->binwidth = init.FCSBinWidth; 
	fcs->TraceLength = cd.TraceLength; 
    fcs->burst_binwidth = ps1_current->BinWidth;
	GetCorrelationParameters(fcs, start, end, init.SumStep);
	GetCorrelationMemory(fcs);
    CalcTime = CrossCorrelateMultiTau(ps1_current, ps2_current, fcs, start, end); 
 }
 else {                                    // reload data and autocorrelate at different binwidth
   	ps1_current = GETPS(set); 
	ps2_current = GETPS(GETOTHER(set));
	strcpy(szBuf, ps1_current->szFilePath);   // 3-17-09
	remove_fext(szBuf);
	sprintf(szFilename, "%sX%s", szBuf, ps2_current->szFilePath);   // 3-17-09
    strcpy(fcs->szFilename, ps1_current->szFilePath);   // 3-17-09 
    fcs->burst_binwidth = ps1_current->BinWidth;
    fcs->binwidth = init.FCSBinWidth; 
	fcs->TraceLength = cd.TraceLength; 
	result = LoadPSWithNewBinWidth(ps1_current->szFilePath, &psA, init.FCSBinWidth) ;  // 3-17-09
	if(result) return;   // file or mem error - trap later
	result = LoadPSWithNewBinWidth(ps2_current->szFilePath, &psB, init.FCSBinWidth) ;  // 3-17-09
	if(result) return;   // file or mem error --trap later
	start = GetPSIndex(ps1_current->ACindex1, ps1_current, &psA); 
	end = GetPSIndex(ps1_current->ACindex2, ps2_current, &psB); 
	GetCorrelationParameters(fcs, start, end, init.SumStep);
	GetCorrelationMemory(fcs);
    CalcTime = CrossCorrelateMultiTau(&psA, &psB, fcs, start, end); 
    sprintf(szBuf, "calc.time = %3.2f sec.", CalcTime); 
    PlotFCSCurve(fcs, szBuf);
    FreePhotonStreamMem(&psA); 
    FreePhotonStreamMem(&psB); 
 }
 StartTime = (double) start * init.FCSBinWidth * US_TO_SEC;
 EndTime = (double) end * init.FCSBinWidth * US_TO_SEC;
 sprintf(szBuf, "CC from %g to %g s,  MT step = %d (calc:%3.2f sec.)", StartTime, EndTime, init.SumStep, CalcTime); 
 PlotFCSCurve(fcs, szBuf);
 if(!bShowResultsWin) ToggleResultsWindow(hwnd)	 ; 
 ARROW_CURSOR;					  
}


#define SUMSTEP_START  2            // usually 2 for in the short lags in the autocorrelation


/******************************************************************************************
  Perform MultiTau correlation
 *******************************************************************************************/
double AutoCorrelateMultiTau(PHOTON_STREAM *ps, AUTOCORRELATION *fcs, DWORD start, DWORD end)
{
 DWORD i, k, j, n, lag, s, endpt, nend, SumStep;
 DWORD *lpFCS;
 double p, denom, BinWidth, BinWidth_ms;
 DWORD dwTime = GetTickCount();

 // reserve working memory
 fcs->dwPtsToCorr = n = end - start + 1; 
 lpFCS =  (DWORD *) GetVirtualMem(sizeof(DWORD) * fcs->dwPtsToCorr);
 if(lpFCS == NULL) return(0.0);  // not able to allocate enough memory

 SumStep = fcs->sumstep; 
 BinWidth = fcs->binwidth;
 BinWidth_ms = BinWidth/1000.0;
 s = SUMSTEP_START * SumStep;
 lag = 1;
 k = 0;
 
 
 // load lpFCS array for summation:
 for(i=0;i<n;i++) lpFCS[i] = (DWORD) ps->counts[i + start];
 
 // calculate the first SUMSTEP_START * SumStep pairs
 while(k < s) {
     endpt = n - lag;	
	 for(i=0; i<endpt; i++) {
	    fcs->ac[k].Go += UInt32x32To64(lpFCS[i], lpFCS[i+lag]);
	    fcs->ac[k].Mi += (__int64) lpFCS[i];
        fcs->ac[k].Mk += (__int64) lpFCS[i+lag];
	 }
     fcs->ac[k].n += endpt; 
	 fcs->ac[k].Lag = lag * BinWidth_ms;   // store as ms
     fcs->ac[k].Gn = (((double) fcs->ac[k].n * fcs->ac[k].Go)/(fcs->ac[k].Mi * fcs->ac[k].Mk)) - 1.0;  
     k++;      // increment ac index counter
     lag++;    // increment lag
 }
 
 // calculate the rest, doubling binwidth every init.SumStep interations 
 s = SumStep;   // reset sumstep 
 p = 1.0;
 while(p <= fcs->FcsMaxPow2) {
	 j = 0;
	 for(i=0;i<n;i+=2) {         // double array
		 lpFCS[j] = lpFCS[i] + lpFCS[i+1]; 
		 j++;
	 }
     n = j;
	 
	 for(j=0;j<s;j++) {         // loop over SumSteps
         lag = s + 1 + j;
		 nend = n - lag;
		 for(i=0;i<nend;i++) {
			  fcs->ac[k].Go += (signed __int64) UInt32x32To64(lpFCS[i], lpFCS[i+lag]);
			  fcs->ac[k].Mi += lpFCS[i];
              fcs->ac[k].Mk += lpFCS[i+lag];
		 }
		 fcs->ac[k].Lag = (double) lag * pow(2.0, p) * BinWidth_ms;   // store as ms
         denom = (double) (fcs->ac[k].Mi * fcs->ac[k].Mk); 
		 if(denom > 0.0) {
	        fcs->ac[k].n += nend; 
		 	fcs->ac[k].Gn = (((double) fcs->ac[k].n * fcs->ac[k].Go)/denom) - 1.0;  
		 }
		 else {
		    fcs->ac[k].Gn = 0.0;
		 }
		 k++;
	 }
     p+=1.0; 
 }

 VirtualFree(lpFCS, 0, MEM_RELEASE);
 lpFCS = NULL;
 dwTime = GetTickCount() - dwTime;
 return((double) dwTime / 1000.0);
}


/******************************************************************************************
  Perform MultiTau cross-correlation
 *******************************************************************************************/
double CrossCorrelateMultiTau(PHOTON_STREAM *ps1, PHOTON_STREAM *ps2, AUTOCORRELATION *fcs, DWORD start, DWORD end)
{

 DWORD i, k, j, n, lag, s, endpt, nend;
 double p, BinWidth, denom, BinWidth_ms; 
 DWORD dwTime = GetTickCount();
 DWORD SumStep; 
 DWORD *lpFCS1, *lpFCS2;

  // reserve working memory
 fcs->dwPtsToCorr = end - start + 1; 
 lpFCS1 =  (DWORD *) GetVirtualMem(sizeof(DWORD) * fcs->dwPtsToCorr);
 lpFCS2 =  (DWORD *) GetVirtualMem(sizeof(DWORD) * fcs->dwPtsToCorr);
 if(lpFCS1 == NULL || lpFCS2 == NULL) return(0.0);    // not able to allocate enough memory

 SumStep = fcs->sumstep; 
 BinWidth = fcs->binwidth;
 BinWidth_ms = BinWidth/1000.0;
 n = end - start + 1;
 s = SUMSTEP_START * SumStep;
 lag = 1;
 k = 0;
  
 // load lpFCS array for summation:
 for(i=0;i<n;i++) {
	 lpFCS1[i] = (DWORD) ps1->counts[i + start];
	 lpFCS2[i] = (DWORD) ps2->counts[i + start];
 }

 // calculate the first SUMSTEP_START * SumStep pairs
 while(k < s) {
     endpt = n - lag;	
	 for(i=0; i<endpt; i++) {
	    fcs->ac[k].Go += ((signed __int64) UInt32x32To64(lpFCS1[i], lpFCS2[i+lag]));
	    fcs->ac[k].Mi += (__int64) lpFCS1[i];
        fcs->ac[k].Mk += (__int64) lpFCS2[i+lag];
	 }
     fcs->ac[k].n = endpt; 
	 fcs->ac[k].Lag = lag * BinWidth_ms;   // store as ms
     fcs->ac[k].Gn = ((double) (fcs->ac[k].n * fcs->ac[k].Go)/(double)(fcs->ac[k].Mi * fcs->ac[k].Mk)) - 1.0;  
     k++;      // increment ac index counter
     lag++;    // increment lag
 }
 

 // calculate the rest, doubling binwidth every init.SumStep interations 
 s = SumStep;   // reset sumstep 
 p = 1.0;
 while(p <= fcs->FcsMaxPow2) {
       j = 0;
	   for(i=0;i<n;i+=2) {       // double arrays
		   lpFCS1[j] = lpFCS1[i] + lpFCS1[i+1]; 
		   lpFCS2[j] = lpFCS2[i] + lpFCS2[i+1];
		   j++;
	   }
       n = j;
	 
	   for(j=0;j<s;j++) {       // loop over init.SumSteps
           lag = s + 1 + j;
		   nend =  n - lag;
		 for(i=0;i<nend;i++) {
			  fcs->ac[k].Go +=  UInt32x32To64(lpFCS1[i], lpFCS2[i+lag]);
	          fcs->ac[k].Mi += (__int64) lpFCS1[i];
              fcs->ac[k].Mk += (__int64) lpFCS2[i+lag];
		 }
         fcs->ac[k].Lag = (double) lag * pow(2.0, p) * BinWidth_ms; // lag in ms
		 denom = (double) (fcs->ac[k].Mi * fcs->ac[k].Mk); 
		 if(denom > 0.0) {
		     fcs->ac[k].n = nend;
		     fcs->ac[k].Gn = (((double) fcs->ac[k].n * fcs->ac[k].Go)/denom) - 1.0;  
		 }
		 else
			 fcs->ac[k].Gn = 0.0;

		 k++;
	 }
     p += 1.0; 
 }

 VirtualFree(lpFCS1, 0, MEM_RELEASE);
 lpFCS1 = NULL;
 VirtualFree(lpFCS2, 0, MEM_RELEASE);
 lpFCS2 = NULL;

 dwTime = GetTickCount() - dwTime;
 return((double) dwTime / 1000.0);
}



/**************************************************************************************
  Analyses the burst data structure to (1) rearrange the start and stop indices to reflect
  the inbetween regions, (2) calculates the between burst intervals, and finds the number 
  of segments greater than the user-specficed minimum time.
  *************************************************************************************/
BOOL SegmentAnalysis(BURST *burst, AUTOCORRELATION *fcs, DWORD numbins)
{
 int i, n, nc, maxindex;
 DWORD s;

 n = (int) burst->numbursts;
 if(n+1 > MAXBUR) n = MAXBUR - 1;
 fcs->burst_binwidth = burst->BinWidth; 
 
 memset(fcs->ValidSeg, 0, MAXBUR);
 memset(fcs->end_index, 0, MAXBUR * sizeof(DWORD));
 memset(fcs->start_index, 0, MAXBUR * sizeof(DWORD));
 memset(fcs->segtimes, 0, MAXBUR * sizeof(double));

 s = burst->End[0];
 fcs->maxinterval = 0.0;
 nc = 0;
 for(i=0;i<n;i++) {
	 fcs->start_index[i] = burst->End[i] - s;
	 fcs->end_index[i] = burst->Start[i];
	 fcs->segtimes[i] = fcs->burst_binwidth * (double) (fcs->end_index[i] - fcs->start_index[i]);
	 if(fcs->segtimes[i] > fcs->mintime) {
		fcs->ValidSeg[i] = 1; 
		nc++;
		if(fcs->segtimes[i] > fcs->maxinterval) {
	       fcs->maxinterval = fcs->segtimes[i];
	       maxindex = i;  // needed for calculating the number of tau's needed for ac
		}
	 }
 }
 
 // fill in the last segment:
 fcs->end_index[n] = numbins-1;
 fcs->start_index[n] = burst->End[n-1];
 fcs->segtimes[n] = fcs->burst_binwidth * (double)(fcs->end_index[n] - fcs->start_index[n]); 
 if(fcs->segtimes[n] > fcs->mintime) {
	fcs->ValidSeg[n] = 1; 
	nc++;
	if(fcs->segtimes[n] > fcs->maxinterval) {
	   fcs->maxinterval = fcs->segtimes[n];
	   maxindex = n;  // needed for calculating the number of tau's needed for ac
	}
 }
 fcs->nave = nc;
 fcs->segmaxpts = (fcs->end_index[maxindex] - fcs->start_index[maxindex] + 1) * (DWORD) (fcs->burst_binwidth/fcs->binwidth);
 fcs->numseg = burst->numbursts + 1;    // correct most of the time, unless a burst starts of ends the PS
 fcs->maxindex = maxindex;
 return(0);
}


void ReSegmentAnalysis(BURST *burst, AUTOCORRELATION *fcs)
{
 int i, n, nc, maxindex;
 
 fcs->maxinterval = 0.0;
 memset(fcs->ValidSeg, 0, MAXBUR);
 nc = 0;
 n = burst->numbursts + 1; 
 for(i=0;i<n;i++) {
	 if(fcs->segtimes[i] > fcs->mintime) {
		fcs->ValidSeg[i] = 1; 
		nc++;
        if(fcs->segtimes[i] > fcs->maxinterval) {
	       fcs->maxinterval = fcs->segtimes[i];
	       maxindex = i;  // needed for calculating the number of tau's needed for ac
		}
	 }
 }
 
 fcs->nave = nc;
 fcs->segmaxpts = (fcs->end_index[maxindex] - fcs->start_index[maxindex] + 1) * (DWORD) (fcs->burst_binwidth/fcs->binwidth);
 fcs->numseg = burst->numbursts + 1;    // correct most of the time, unless a burst starts of ends the PS
 fcs->maxindex = maxindex;
}

/**************************************************************************************
  Autocorrelates and averages the photon count data between bursts
 *************************************************************************************/
BOOL AutocorrelateSegments(AUTOCORRELATION *fcs, int set)
{
 char   szBuf[256];
 int    result; 
 DWORD  i, start, end;
 PHOTON_STREAM ps, *ps_current; 
 
 HOURGLASS_CURSOR;
 strcpy(szBuf, "Segmented AC"); 

 fcs->DataType = cd.DataType = SEGFCS_DATA;
 ps_current = GETPS(set);     // get the filename of the open PS
 strcpy(szBuf, ps_current->szFilePath);    // 3-17-09 
 result = LoadPSWithNewBinWidth(szBuf, &ps, init.FCSBinWidth) ;
 if(result) return(MEM_ERROR);   // file error
 fcs->binwidth = init.FCSBinWidth; 
 fcs->TraceLength = cd.TraceLength; 
 GetCorrelationParameters(fcs, fcs->start_index[fcs->maxindex], fcs->end_index[fcs->maxindex], init.SumStep);
 GetCorrelationMemory(fcs);
 fcs->sumstep = init.SumStep; 
 for(i=0;i<fcs->nave;i++) {
     if(fcs->ValidSeg[i]) {
	    GetCorrelationParameters(fcs, fcs->start_index[i], fcs->end_index[i], init.SumStep);
	    start = GetPSIndex(fcs->start_index[i], ps_current, &ps); 
	    end = GetPSIndex(fcs->end_index[i], ps_current, &ps); 
	    AutoCorrelateMultiTau(&ps, fcs, start, end); 
	    PlotFCSCurve(fcs, szBuf);
	}
 }
 FreePhotonStreamMem(&ps); 
 ARROW_CURSOR;					  
 return(FALSE);
}






  


