/*****************************************************************************
  BurstMem.c  - memory management functions
  Version 1.0
  July 2008 WRZ
*******************************************************************************/
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "resource.h"
#include "Burst.h"

/********************************************************************************
  Gets virtual memory handle 
 **********************************************************************************/ 
LPVOID GetVirtualMem(DWORD dwSize) // dwSize is the number of bytes needed
{
  DWORD dwPageSize, dwNumPages;
  SYSTEM_INFO SystemInfo;

  GetSystemInfo(&SystemInfo);     // get system buffer size
  dwPageSize = SystemInfo.dwPageSize;
  dwNumPages = (dwSize % dwPageSize)?(1 + dwSize/dwPageSize):(dwSize/dwPageSize);
  return(VirtualAlloc(NULL, dwPageSize * dwNumPages, MEM_COMMIT, PAGE_READWRITE));
} 


/**************************************************************************************
 Free memory from PHOTON_STREAM structure
 ************************************************************************************/
int FreePhotonStreamMem(PHOTON_STREAM *ps)
{
  if(ps->counts) {
	    VirtualFree(ps->counts, 0, MEM_RELEASE);  
        ps->counts = NULL;
		ps->nbins = 0;
        ps->BaseBinWidth = 0.0;
        ps->BinWidth = 0.0;
        ps->CPS = 0.0;
        ps->FileType = UNSUPPORTED;
        ps->max = 0;
        ps->NumPhotons = 0;
        ps->szFilename[0] = '\0';
        ps->TraceLength = 0.0;
		ps->SGFilterPasses = 0;
		return(NO_ERRORS);
  }
  return(MEM_ERROR);
}


/***********************************************************************************
  Reserve memory for multitau autocorrelation
 **********************************************************************************/
BOOL GetCorrelationMemory(AUTOCORRELATION *fcs)
{
  fcs->ac = (AUTOCORR_DATA *) calloc(fcs->ntau,  sizeof(AUTOCORR_DATA));
  if(fcs->ac == NULL) return(MEM_ERROR);
  return(NO_ERRORS); 
}

/***********************************************************************************
  Calculate the number of data points and other info needed for the 
  photon count segment being multitau correlated 
 **********************************************************************************/
void GetCorrelationParameters(AUTOCORRELATION *fcs, DWORD start, DWORD end, WORD SumStep)
{
  int numpts; 

  fcs->binwidth = init.FCSBinWidth;    // this may have been changed by user via the dialog box
  fcs->sumstep = SumStep;              // this may have been changed by user via the dialog box
  numpts = end - start + 1; 
  fcs->FcsMaxPow2 = (double) floor(log(FCS_FRACTION * (double) numpts / (2.0 * (double) SumStep))/LN2); 
  fcs->ntau = (DWORD) ((double) SumStep * (fcs->FcsMaxPow2 + 2));
}


/******************************************************************************************
 Reset AUTOCORRELATION struct and free allocated memory
 *******************************************************************************************/
void FreeCorrelationMemory(AUTOCORRELATION *fcs)
{
 fcs->burst_binwidth = 0.0;
 fcs->binwidth = 0.0;
 fcs->dwPtsToCorr = 0;
 fcs->FcsMaxPow2 = 0;
 fcs->maxinterval = 0;
 fcs->ntau = 0;
 fcs->nave = 0;
 fcs->sumstep = 8;
 fcs->szFilename[0] = '\0';
 free(fcs->ac); 
}



/******************************************************************************************
   Free PCH memory
 *******************************************************************************************/
void FreePCHMemory(PC_HIST *pch)
{
 pch->n = 0;
 free(pch->x);
 free(pch->y);
 pch->xmax = 0.0;
 pch->ymax = 0.0;
}

/******************************************************************************************
 Reserve memory for the burst analysis routines
 *******************************************************************************************/
BOOL GetBurstMem(BURST *burst, PHOTON_STREAM *ps)
{
  DWORD dwNumBinsAboveTrigger; 
  
  // calculate dwAnalysisSize based on number of values above trigger
  dwNumBinsAboveTrigger = 2 * NumberBinsAbove(burst->TriggerCounts - 1, ps->counts, ps->nbins);   // double just to be sure
  if(dwNumBinsAboveTrigger == 0) return(NO_BINS_ABOVE_TRIGGER);
  burst->dwAnalysisSize = dwNumBinsAboveTrigger * 8;   // size in bytes to allocate for each burst parameter array
  burst->dwAnalysisPts = dwNumBinsAboveTrigger;        // maximum array size
    
  if((burst->Burst = (double *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  if((burst->Width = (double *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  if((burst->Interval = (double *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  if((burst->Peak = (double *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  if((burst->Time = (double *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  if((burst->Start = (DWORD *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  if((burst->End = (DWORD *) GetVirtualMem(burst->dwAnalysisSize)) == NULL) return(MEM_ERROR);
  return (NO_ERRORS);
}

/******************************************************************************************
 Frees virtual arrays in BURST structure
 *******************************************************************************************/
void FreeBurstMem(BURST *burst) 
{
 int i;
 VirtualFree(burst->Burst, 0, MEM_RELEASE);
 burst->Burst = NULL;
 
 VirtualFree(burst->Width, 0, MEM_RELEASE);
 burst->Width = NULL;
 
 VirtualFree(burst->Interval, 0, MEM_RELEASE);
 burst->Interval = NULL;
 
 VirtualFree(burst->Peak, 0, MEM_RELEASE);
 burst->Peak = NULL;
 
 VirtualFree(burst->Time, 0, MEM_RELEASE);
 burst->Time = NULL;
 
 VirtualFree(burst->Start, 0, MEM_RELEASE);
 burst->Start = NULL;
 
 VirtualFree(burst->End, 0, MEM_RELEASE);
 burst->End = NULL;
 
 burst->numbursts = 0;
 burst->BackGrdCounts = 0;
 burst->BinWidth = 0.0;
 burst->BkMean = 0.0;
 burst->BkSD = 0.0;
 burst->dwAnalysisPts = 0;
 burst->dwAnalysisSize = 0;

 for(i=0;i<3;i++) WGFreeDataSet(pd[i].hData, TRUE);
}


/************************************************************************
   Error handling functions
 ***********************************************************************/
void DisplayErrorMessage(void)
{
 LPVOID lpMsgBuf;

 FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
               NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL);
 MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
 LocalFree( lpMsgBuf );   // Free the buffer.
}
/************************************************************************
  Displays an error messagebox if there is a virtual memory error
 ***********************************************************************/

void VirtualMemError(char *szStr)
{
 LPVOID lpMsgBuf;
 char szBuf[128]; 

 wsprintf(szBuf,"Unrecoverable error while freeing %s virtual memory array", szStr); 
 FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS,
               NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL);
 MessageBox( NULL, (LPCTSTR)lpMsgBuf, szBuf, MB_OK | MB_ICONINFORMATION );
 LocalFree( lpMsgBuf );   
}
